@@ -1,6 +1,6 @@
 /*
  * SGMLOP
- * $Id: sgmlop.c,v 1.10 2001/12/30 23:42:40 loewis Exp $
+ * $Id: sgmlop.c,v 1.14 2002/07/31 06:04:31 loewis Exp $
  *
  * The sgmlop accelerator module
  *
@@ -61,6 +61,7 @@
 #if (PY_MAJOR_VERSION == 1 && PY_MINOR_VERSION > 5) || (PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION < 2)
 /* In Python 1.6, 2.0 and  2.1, disabling Unicode was not possible. */
 #define Py_USING_UNICODE
+#define PyUnicode_GetMax()  (0xffff)
 #endif
 
 #ifdef SGMLOP_UNICODE_SUPPORT
@@ -1140,9 +1141,15 @@ fastfeed(FastSGMLParserObject* self)
                     goto eol;
             }
             b = t = p;
-            while (ISALNUM(*p) || *p == '.')
-                if (++p >= end)
-                    goto eol;
+            if (self->xml) {
+                while (ISALNUM(*p) || *p == '.' || *p == '-' || *p == '_' || *p == ':')
+                    if (++p >= end)
+                        goto eol;
+            } else {
+                while (ISALNUM(*p) || *p == '.')
+                    if (++p >= end)
+                        goto eol;
+            }
             e = p;
             if (*p == ';')
                 p++;
@@ -1222,14 +1229,13 @@ fastfeed(FastSGMLParserObject* self)
             }
             else {
                 /* fallback: handle charref's as data */
-                int ch;
+                int ch = 0;
                 CHAR_T *p;
-                ch = 0;
                 if (*b == 'x') {
                     for (p = b+1; p < e; p++)
                         ch = ch*16 + *p - (*p > 'F' ? 
-					   'a'-1 :(*p > '9' ? 
-						   'A'-1 : '0'));
+					   'a'-10 :(*p > '9' ? 
+						   'A'-10 : '0'));
                 } else {
                     for (p = b; p < e; p++)
                         ch = ch*10 + *p - '0';
@@ -1238,6 +1244,13 @@ fastfeed(FastSGMLParserObject* self)
 		if (self->unicode) {
 		    PyObject *res;
 		    Py_UNICODE uch = ch;
+			 int maxunicode = PyUnicode_GetMax();
+
+		    if (ch > maxunicode) {
+			PyErr_Format(PyExc_ValueError,
+			    "character reference &#x%x; exceeds sys.maxunicode (0x%x)", ch, maxunicode);
+			return -1;
+		    }
 		    res = PyObject_CallFunction(self->handle_data,
 						"u#", &uch, 1);
 		    if (!res)
@@ -1247,11 +1260,11 @@ fastfeed(FastSGMLParserObject* self)
 #endif
 		{
 		    char nch;
-		    if (ch > 128) {
-			/* XXX: should utf-8 encode here. */
-			PyErr_SetString(PyExc_ValueError, 
-					"character reference too large");
-			return -1;
+		    if (ch >= 128) {
+			/* XXX: should utf-8 encode here for XML; can't do anything for SGML. */
+		     PyErr_Format(PyExc_ValueError, 
+				     "character reference &#x%x; exceeds ASCII range", ch);
+		     return -1;
 		    }
 		    nch = ch;
 		    if (callHandleData(self, &nch, 1))
@@ -1458,7 +1471,7 @@ fetchEncoding(FastSGMLParserObject* self, const CHAR_T* data, int len)
     }
     strncpy(self->encoding, found, data-found);
     self->encoding[data-found] = '\0';
-    printf("'%s'\n", self->encoding);
+    /*printf("'%s'\n", self->encoding);*/
     return 0;
 }
 
